package collector

import (
	"errors"
	"fmt"
	"testing"
	"time"

	"github.com/stretchr/testify/require"
	v1 "github.com/zalando-incubator/kube-metrics-adapter/pkg/apis/zalando.org/v1"
	autoscalingv2 "k8s.io/api/autoscaling/v2beta2"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
)

const hHMMFormat = "15:04"

type schedule struct {
	kind      string
	date      string
	startTime string
	days      []v1.ScheduleDay
	timezone  string
	duration  int
	value     int
}

func TestScalingScheduleCollector(t *testing.T) {
	nowRFC3339 := "2009-11-10T23:00:00+01:00" // +01:00 is Berlin timezone (default)
	nowTime, _ := time.Parse(time.RFC3339, nowRFC3339)
	nowWeekday := v1.TuesdaySchedule

	// now func always return in UTC for test purposes
	utc, _ := time.LoadLocation("UTC")
	uTCNow := nowTime.In(utc)
	uTCNowRFC3339 := uTCNow.Format(time.RFC3339)
	now := func() time.Time {
		return uTCNow
	}

	for _, tc := range []struct {
		msg           string
		schedules     []schedule
		expectedValue int
		err           error
	}{
		{
			msg: "Return the right value for one time config",
			schedules: []schedule{
				{
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the right value for one time config - ten minutes after starting a 15 minutes long schedule",
			schedules: []schedule{
				{
					date:     nowTime.Add(-time.Minute * 10).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the right value for one time config - 30 seconds before ending",
			schedules: []schedule{
				{
					date:     nowTime.Add(-time.Minute * 15).Add(time.Second * 30).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the default value (0) for one time config - 30 seconds after",
			schedules: []schedule{
				{
					date:     nowTime.Add(-time.Minute * 45).Add(-time.Second * 30).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 45,
					value:    100,
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the default value (0) for one time config not started yet (20 minutes before)",
			schedules: []schedule{
				{
					date:     nowTime.Add(time.Minute * 20).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 65,
					value:    100,
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the default value (0) for one time config that ended (20 minutes after now for a 15 minutes long schedule)",
			schedules: []schedule{
				{
					date:     nowTime.Add(-time.Minute * 20).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return error for one time config not in RFC3339 format",
			schedules: []schedule{
				{
					date:     nowTime.Add(-time.Minute * 20).Format(time.RFC822),
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			err: ErrInvalidScheduleDate,
		},
		{
			msg: "Return the right value for two one time config",
			schedules: []schedule{
				{
					// 20 minutes after now for a 15 minutes long schedule
					date:     nowTime.Add(-time.Minute * 20).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
				{
					// starting now
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the biggest value for multiples one time configs (all starting now)",
			schedules: []schedule{
				{
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    100,
				},
				{
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    120,
				},
				{
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    110,
				},
			},
			expectedValue: 120,
		},
		{
			msg: "Return the right value for a repeating schedule",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the right value for a repeating schedule - 5 minutes after started",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Add(-5 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the right value for a repeating schedule - 5 minutes before ending",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Add(-10 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return the default value (0) for a repeating schedule in the wrong day of the week",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Format(hHMMFormat),
					days:      []v1.ScheduleDay{v1.MondaySchedule}, // Not nowWeekday
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the right value for a repeating schedule in the right timezone",
			schedules: []schedule{
				{
					kind:     "Repeating",
					duration: 15,
					value:    100,
					// Sao Paulo is -3 hours from Berlin
					startTime: nowTime.Add(-time.Hour * 3).Format(hHMMFormat),
					timezone:  "America/Sao_Paulo",
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return an error if the start time is not in the format HH:MM",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: "15h25min",
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			err: ErrInvalidScheduleStartTime,
		},
		{
			msg: "Return the right value for a repeating schedule in the right timezone even in the day after it",
			schedules: []schedule{
				{
					kind:     "Repeating",
					duration: 15,
					value:    100,
					// Tokyo is +8 hours from Berlin
					startTime: nowTime.Add(time.Hour * 8).Format(hHMMFormat),
					timezone:  "Asia/Tokyo",
					// It's Wednesday in Japan
					days: []v1.ScheduleDay{v1.WednesdaySchedule},
				},
			},
			expectedValue: 100,
		},
		{
			msg: "Return default (0) for a repeating schedule in the right timezone in the day after it",
			schedules: []schedule{
				{
					kind:     "Repeating",
					duration: 15,
					value:    100,
					// Tokyo is +8 hours from Berlin
					startTime: nowTime.Add(time.Hour * 8).Format(hHMMFormat),
					timezone:  "Asia/Tokyo",
					// It's Wednesday in Japan
					days: []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the default value (0) for a repeating schedule in the right day of the week but five minutes too early",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Add(5 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the default value (0) for a repeating schedule in the right day of the week but five minutes too late (schedule started 20 minutes ago and lasted 15.)",
			schedules: []schedule{
				{
					kind:      "Repeating",
					duration:  15,
					value:     100,
					startTime: nowTime.Add(-20 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 0,
		},
		{
			msg: "Return the biggest value for multiple repeating schedules",
			schedules: []schedule{
				{
					// in time schedule - 30 minutes before
					kind:      "Repeating",
					duration:  60,
					value:     110,
					startTime: nowTime.Add(-30 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// in time schedule - 1 minute before ending
					// biggest value
					kind:      "Repeating",
					duration:  6,
					value:     120,
					startTime: nowTime.Add(-5 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// in time schedule - 10 minute after starting
					kind:      "Repeating",
					duration:  35,
					value:     100,
					startTime: nowTime.Add(-10 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 120,
		},
		{
			msg: "Return the biggest value for multiple repeating schedules - 1 minute too late for 120",
			schedules: []schedule{
				{
					// in time schedule - 30 minutes before
					// biggest value
					kind:      "Repeating",
					duration:  60,
					value:     110,
					startTime: nowTime.Add(-30 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// not in time schedule - 1 minute too late
					kind:      "Repeating",
					duration:  4,
					value:     120,
					startTime: nowTime.Add(-5 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// in time schedule - 10 minute after starting
					kind:      "Repeating",
					duration:  35,
					value:     100,
					startTime: nowTime.Add(-10 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
			},
			expectedValue: 110,
		},
		{
			msg: "Return the biggest value for multiple repeating and oneTime configs",
			schedules: []schedule{
				{
					// in time schedule - 30 minutes before
					// biggest value
					kind:      "Repeating",
					duration:  60,
					value:     110,
					startTime: nowTime.Add(-30 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// not in time schedule - 1 minute too late
					kind:      "Repeating",
					duration:  4,
					value:     130,
					startTime: nowTime.Add(-5 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// in time schedule - 10 minutes after starting
					kind:      "Repeating",
					duration:  35,
					value:     100,
					startTime: nowTime.Add(-10 * time.Minute).Format(hHMMFormat),
					days:      []v1.ScheduleDay{nowWeekday},
				},
				{
					// in time schedule
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    90,
				},
				{
					// invalid schedule - start tomorrow
					date:     nowTime.Add(24 * time.Hour).Format(time.RFC3339),
					kind:     "OneTime",
					duration: 15,
					value:    140,
				},
				{
					date:     nowRFC3339,
					kind:     "OneTime",
					duration: 15,
					value:    110,
				},
			},
			expectedValue: 110,
		},
	} {
		t.Run(tc.msg, func(t *testing.T) {
			scalingScheduleName := "my_scaling_schedule"
			namespace := "default"

			schedules := getSchedules(tc.schedules)
			store := newMockStore(scalingScheduleName, namespace, schedules)
			plugin, err := NewScalingScheduleCollectorPlugin(store, now)
			require.NoError(t, err)

			clusterStore := newClusterMockStore(scalingScheduleName, schedules)
			clusterPlugin, err := NewClusterScalingScheduleCollectorPlugin(clusterStore, now)
			require.NoError(t, err)

			clusterStoreFirstRun := newClusterMockStoreFirstRun(scalingScheduleName, schedules)
			clusterPluginFirstRun, err := NewClusterScalingScheduleCollectorPlugin(clusterStoreFirstRun, now)
			require.NoError(t, err)

			hpa := makeScalingScheduleHPA(namespace, scalingScheduleName)

			configs, err := ParseHPAMetrics(hpa)
			require.NoError(t, err)
			require.Len(t, configs, 2)

			collectorFactory := NewCollectorFactory()
			err = collectorFactory.RegisterObjectCollector("ScalingSchedule", "", plugin)
			require.NoError(t, err)
			err = collectorFactory.RegisterObjectCollector("ClusterScalingSchedule", "", clusterPlugin)
			require.NoError(t, err)
			collectorFactoryFirstRun := NewCollectorFactory()
			err = collectorFactoryFirstRun.RegisterObjectCollector("ClusterScalingSchedule", "", clusterPluginFirstRun)
			require.NoError(t, err)

			collector, err := collectorFactory.NewCollector(hpa, configs[0], 0)
			require.NoError(t, err)
			collector, ok := collector.(*ScalingScheduleCollector)
			require.True(t, ok)

			clusterCollector, err := collectorFactory.NewCollector(hpa, configs[1], 0)
			require.NoError(t, err)
			clusterCollector, ok = clusterCollector.(*ClusterScalingScheduleCollector)
			require.True(t, ok)

			clusterCollectorFirstRun, err := collectorFactoryFirstRun.NewCollector(hpa, configs[1], 0)
			require.NoError(t, err)
			clusterCollectorFirstRun, ok = clusterCollectorFirstRun.(*ClusterScalingScheduleCollector)
			require.True(t, ok)

			checkCollectedMetrics := func(t *testing.T, collected []CollectedMetric, resourceType string) {
				if tc.err != nil {
					require.Equal(t, tc.err, err)
				} else {
					require.NoError(t, err, "failed to collect %s metrics: %v", resourceType, err)
					require.Len(t, collected, 1, "the number of metrics returned is not 1")
					require.EqualValues(t, tc.expectedValue, collected[0].Custom.Value.Value(), "the returned metric is not expected value")
					require.EqualValues(t, autoscalingv2.ObjectMetricSourceType, collected[0].Type)
					require.EqualValues(t, scalingScheduleName, collected[0].Custom.DescribedObject.Name)
					require.EqualValues(t, namespace, collected[0].Custom.DescribedObject.Namespace)
					require.EqualValues(t, "zalando.org/v1", collected[0].Custom.DescribedObject.APIVersion)
					require.EqualValues(t, resourceType, collected[0].Custom.DescribedObject.Kind)
					require.EqualValues(t, uTCNowRFC3339, collected[0].Custom.Timestamp.Time.Format(time.RFC3339))
					require.EqualValues(t, collected[0].Custom.Metric.Name, scalingScheduleName)
					require.EqualValues(t, namespace, collected[0].Namespace)
				}
			}

			collected, err := collector.GetMetrics()
			checkCollectedMetrics(t, collected, "ScalingSchedule")

			clusterCollected, err := clusterCollector.GetMetrics()
			checkCollectedMetrics(t, clusterCollected, "ClusterScalingSchedule")

			clusterCollectedFirstRun, err := clusterCollectorFirstRun.GetMetrics()
			checkCollectedMetrics(t, clusterCollectedFirstRun, "ClusterScalingSchedule")
		})
	}
}

func TestScalingScheduleObjectNotPresentReturnsError(t *testing.T) {
	store := mockStore{
		make(map[string]interface{}),
		getByKeyFn,
	}
	plugin, err := NewScalingScheduleCollectorPlugin(store, time.Now)
	require.NoError(t, err)

	clusterStore := mockStore{
		make(map[string]interface{}),
		getByKeyFn,
	}
	clusterPlugin, err := NewClusterScalingScheduleCollectorPlugin(clusterStore, time.Now)
	require.NoError(t, err)

	hpa := makeScalingScheduleHPA("namespace", "scalingScheduleName")

	configs, err := ParseHPAMetrics(hpa)
	require.NoError(t, err)
	require.Len(t, configs, 2)

	collectorFactory := NewCollectorFactory()
	err = collectorFactory.RegisterObjectCollector("ScalingSchedule", "", plugin)
	require.NoError(t, err)
	err = collectorFactory.RegisterObjectCollector("ClusterScalingSchedule", "", clusterPlugin)
	require.NoError(t, err)

	collector, err := collectorFactory.NewCollector(hpa, configs[0], 0)
	require.NoError(t, err)
	collector, ok := collector.(*ScalingScheduleCollector)
	require.True(t, ok)

	clusterCollector, err := collectorFactory.NewCollector(hpa, configs[1], 0)
	require.NoError(t, err)
	clusterCollector, ok = clusterCollector.(*ClusterScalingScheduleCollector)
	require.True(t, ok)

	_, err = collector.GetMetrics()
	require.Error(t, err)
	require.Equal(t, ErrScalingScheduleNotFound, err)

	_, err = clusterCollector.GetMetrics()
	require.Error(t, err)
	require.Equal(t, ErrClusterScalingScheduleNotFound, err)

	var invalidObject struct{}

	store.d["namespace/scalingScheduleName"] = invalidObject
	clusterStore.d["scalingScheduleName"] = invalidObject

	_, err = collector.GetMetrics()
	require.Error(t, err)
	require.Equal(t, ErrNotScalingScheduleFound, err)

	_, err = clusterCollector.GetMetrics()
	require.Error(t, err)
	require.Equal(t, ErrNotClusterScalingScheduleFound, err)
}

func TestReturnsErrorWhenStoreDoes(t *testing.T) {
	store := mockStore{
		make(map[string]interface{}),
		func(d map[string]interface{}, key string) (item interface{}, exists bool, err error) {
			return nil, true, errors.New("Unexpected store error")
		},
	}

	plugin, err := NewScalingScheduleCollectorPlugin(store, time.Now)
	require.NoError(t, err)

	clusterPlugin, err := NewClusterScalingScheduleCollectorPlugin(store, time.Now)
	require.NoError(t, err)

	hpa := makeScalingScheduleHPA("namespace", "scalingScheduleName")
	configs, err := ParseHPAMetrics(hpa)
	require.NoError(t, err)
	require.Len(t, configs, 2)

	collectorFactory := NewCollectorFactory()
	err = collectorFactory.RegisterObjectCollector("ScalingSchedule", "", plugin)
	require.NoError(t, err)
	err = collectorFactory.RegisterObjectCollector("ClusterScalingSchedule", "", clusterPlugin)
	require.NoError(t, err)

	collector, err := collectorFactory.NewCollector(hpa, configs[0], 0)
	require.NoError(t, err)
	collector, ok := collector.(*ScalingScheduleCollector)
	require.True(t, ok)

	clusterCollector, err := collectorFactory.NewCollector(hpa, configs[1], 0)
	require.NoError(t, err)
	clusterCollector, ok = clusterCollector.(*ClusterScalingScheduleCollector)
	require.True(t, ok)

	_, err = collector.GetMetrics()
	require.Error(t, err)

	_, err = clusterCollector.GetMetrics()
	require.Error(t, err)
}

type mockStore struct {
	d          map[string]interface{}
	getByKeyFn func(d map[string]interface{}, key string) (item interface{}, exists bool, err error)
}

func (m mockStore) GetByKey(key string) (item interface{}, exists bool, err error) {
	return m.getByKeyFn(m.d, key)
}

func getByKeyFn(d map[string]interface{}, key string) (item interface{}, exists bool, err error) {
	item, exists = d[key]
	return item, exists, nil
}

func newMockStore(name, namespace string, schedules []v1.Schedule) mockStore {
	return mockStore{
		map[string]interface{}{
			fmt.Sprintf("%s/%s", namespace, name): &v1.ScalingSchedule{
				ObjectMeta: metav1.ObjectMeta{
					Name: name,
				},
				Spec: v1.ScalingScheduleSpec{
					Schedules: schedules,
				},
			},
		},
		getByKeyFn,
	}
}

func newClusterMockStore(name string, schedules []v1.Schedule) mockStore {
	return mockStore{
		map[string]interface{}{
			name: &v1.ClusterScalingSchedule{
				ObjectMeta: metav1.ObjectMeta{
					Name: name,
				},
				Spec: v1.ScalingScheduleSpec{
					Schedules: schedules,
				},
			},
		},
		getByKeyFn,
	}
}

// The cache.Store returns the v1.ClusterScalingSchedule items as
// v1.ScalingSchedule when it first lists it. When it's update it
// asserts it correctly to the v1.ClusterScalingSchedule type.
func newClusterMockStoreFirstRun(name string, schedules []v1.Schedule) mockStore {
	return mockStore{
		map[string]interface{}{
			name: &v1.ScalingSchedule{
				ObjectMeta: metav1.ObjectMeta{
					Name: name,
				},
				Spec: v1.ScalingScheduleSpec{
					Schedules: schedules,
				},
			},
		},
		getByKeyFn,
	}
}

func getSchedules(schedules []schedule) (result []v1.Schedule) {
	for _, schedule := range schedules {
		switch schedule.kind {
		case string(v1.OneTimeSchedule):
			date := v1.ScheduleDate(schedule.date)
			result = append(result,
				v1.Schedule{
					Type:            v1.OneTimeSchedule,
					Date:            &date,
					DurationMinutes: schedule.duration,
					Value:           schedule.value,
				},
			)
		case string(v1.RepeatingSchedule):
			period := v1.SchedulePeriod{
				StartTime: schedule.startTime,
				Days:      schedule.days,
				Timezone:  schedule.timezone,
			}
			result = append(result,
				v1.Schedule{
					Type:            v1.RepeatingSchedule,
					Period:          &period,
					DurationMinutes: schedule.duration,
					Value:           schedule.value,
				},
			)
		}
	}
	return
}

func makeScalingScheduleHPA(namespace, scalingScheduleName string) *autoscalingv2.HorizontalPodAutoscaler {
	return &autoscalingv2.HorizontalPodAutoscaler{
		ObjectMeta: metav1.ObjectMeta{Namespace: namespace},
		Spec: autoscalingv2.HorizontalPodAutoscalerSpec{
			ScaleTargetRef: autoscalingv2.CrossVersionObjectReference{
				Kind: "Deployment",
				Name: "Application",
			},
			Metrics: []autoscalingv2.MetricSpec{
				{
					Type: autoscalingv2.ObjectMetricSourceType,
					Object: &autoscalingv2.ObjectMetricSource{
						DescribedObject: autoscalingv2.CrossVersionObjectReference{
							Name:       scalingScheduleName,
							APIVersion: "zalando.org/v1",
							Kind:       "ScalingSchedule",
						},
						Metric: autoscalingv2.MetricIdentifier{
							Name: scalingScheduleName,
						},
					},
				}, {
					Type: autoscalingv2.ObjectMetricSourceType,
					Object: &autoscalingv2.ObjectMetricSource{
						DescribedObject: autoscalingv2.CrossVersionObjectReference{
							Name:       scalingScheduleName,
							APIVersion: "zalando.org/v1",
							Kind:       "ClusterScalingSchedule",
						},
						Metric: autoscalingv2.MetricIdentifier{
							Name: scalingScheduleName,
						},
					},
				},
			},
		},
	}
}
